# -*- coding: utf-8 -*-
#Created on Tue Aug 27 11:05:54 2013
#
#@author: gillis

#
# PyRobotran
# Copyright 2013, Universite catholique de Louvain, Belgium
# All rights reserved
# 
# 
# The PyRobotran software has been developed by Thomas Gillis (Louvain School of engineering)
# and Nicolas Docquier (Center for Research in Mechatronics, institute of MEchanics, Materials and 
# Civil engineering) for the Université catholique de Louvain (UCL).
#
# PyRobotran is a free software: you can redistribute it and/or modify it under the terms of 
# the BSD 3-Clause License as published below.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted
# provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright notice, 
#     this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice, 
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#   * Neither the name of the Université catholique de Louvain nor the names of
#     its contributors may be used to endorse or promote products derived from 
#     this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#
# For any use in commercial products or questions about licensing, you can contact
#  - the Center for Research in Mechatronics at info@cerem.be
#  - the Louvain Technology Transfer Office of UCL at ltto@uclouvain.be
#



import numpy as np



class Coord_part:
    """
    Class of the coordinate partionning
    
    Take care: By convention index of lists begins to 0 and index of arrays begins to 1s
    
    
    :param mbs: the mbs_data stucture of the system
    :param is_row_perm: the value is True if we can permutate the rows. False if not
    :param row_perm: a list of the permutation of the row
    :param col_perm: a list of permutation of the columns
    """
    
    def __init__(self,mbs,is_row_perm=1):
        self.mbs=mbs
        self.is_row_perm=is_row_perm
        self.irk=[]
        self.row_perm=[]
        self.col_perm=[]

    
    
    
    def run(self):
        """
        Function to run the coordinate partitionning. This function firstly compute the partition
        when the system is open. Secondly it try to close the system (mbs.clos_geo function) and
        finally it recompute the coordinate partition to imporve it and to correct it if neccesary
        
        """
        self.mbs.donePart=True        
        
        if self.mbs.Ncons == 0:
            print ('No need to partition this system! No constraints')
            return
        
        #truc de variable commandée dans matlab????? pq on fait ça??? pq on met a zero?
        
        #first partitionning: open system
        ind_u_save=list(self.mbs.qu)#backup of the wanted independant coordinates
        self.coord_part(1)
        
        
        #Try to close the system
        self.mbs.close_geo()
        
        # we reset the wanted independant coordinates
        self.mbs.set_qu(ind_u_save)
        #check again partitionning with the closed system
        self.coord_part(2) #we could have set 0 but its used like a boolean, so we don't care
        
        # remise des tailles: nqu, nqv, ... ==> not necessary if we use set_qu, set_qv, ...!
        

    def coord_part(self,first_part):
        """
        Function to partition the coordinates
        
        :param ind_qu_des: list of indices the user want to be independant
        :param ind_c: list of driven coordinates
        :param is_row_perm: flag to allow lignes permutation
        :param first_part: =1 if this is the first partionning, =0 if not
        
        """

        # computing real rank of the Jacobian:
        qsafe=self.mbs.q.copy()
        self.mbs.q[:]=np.random.random(np.size(self.mbs.q))
        self.mbs.calc_hJac()
        real_rank=np.linalg.matrix_rank(self.mbs.J)
        self.mbs.q[:]=qsafe[:]
        
        # partionning proposed (des = desired)
        ind_uc_des=self.mbs.qu + self.mbs.qc #There are LISTS!!!
        ind_uc_des.sort()        
        
        
        ind_tot=range(1,self.mbs.njoint+1)
        ind_vu=self.difference(ind_tot,ind_uc_des)  # Homemade function. Another way???
        
        
        #shaking the system: only for the first partionning
        if first_part:
            for i in range(0,len(ind_vu)):
                if np.abs(self.mbs.q[ind_vu[i]]) < 1e-6:
                    self.mbs.q[ind_vu[i]] = 1e-5*np.random.random()
        
        # LU factorisation
        self.mbs.calc_hJac() #computing the jacobian
        
        self.LUtot(self.mbs.J[1:,ind_vu])
        
        
        
        
        #check the user choice
        if self.irk < real_rank:  #non-valid choice
#            print 'impossible'
            raise Exception('coord_part: wrong choice for the independent variable set - Jv matrix singular')
        else: #valid choice: we replace the indices
            # passage des listes en np array pour avoir plus simple pour les sublister...
            
            
            ind_vu=np.array(ind_vu)
            ind_aux=ind_vu[self.col_perm] #numpy array
            
            # update qv
            self.mbs.reset_qu_qv(self.mbs.qu+(ind_aux[self.irk:].tolist()), ind_aux[0:self.irk].tolist())# le vecteur aux étant bien un joli vecteur sans zéro bizarre devant,
            # on doit donc commencer a l'indice 0 et s'arreter à irk, non compris!
            
            # update qu
            if len(ind_aux) > self.irk:
                self.mbs.set_qu(self.mbs.qu+(ind_aux[self.irk:].tolist()))
            
            
            # on remet tout en place
            self.mbs.q[self.mbs.qu]=qsafe[self.mbs.qu]
                        
            
        
        # asignation of the new index
        self.row_perm=np.array(self.row_perm) #on le met en numpy pour pouvoir jouer avec les indices
        self.row_perm+=1 #on augmente de 1 partout car ça va aller pour des tableau numpy (h)
#        self.row_perm.sort()
        
        if self.irk == self.mbs.J.shape[0]: #jac.shape= (number line, number row)
            self.mbs.reset_hu_hv(self.row_perm.tolist(), [])
        else:
            self.mbs.reset_hu_hv(self.row_perm[0:self.irk].tolist(), self.row_perm[self.irk:].tolist())
        
#        hu should be sorted ? not necessarily
#        self.mbs.hu.sort()
#        self.mbs.hv.sort()
        
    

    def difference(self,list1,list2):
        
        result=[]        
        for elem in list1:
            if elem not in list2:
                result.append(elem)
        
        return result
        



    def LUtot(self,a):
        """ Compute the LU factorisation of the a matrix. It computes the best row
        permutation and the best line permutation. The aim is to have the best conditionned matrix
        in the Newton Raphson resoltuion.

        :param a: the matrix for which we want to determine the best row-line permutations
        
        """
        #initialisation
        (nl,nc)=np.shape(a) #la matrice arrive sans les lignes de 0!
        epsilon=1e-15
        self.irk=0
        ipl=[]
        ipc=[]
        
        
        for ite in range(0,nl):
            
            if self.is_row_perm:
#                ilmax=nl
                rangel=range(ite,nl)
            else:
#                ilmax=ite
                rangel=[ite]
            
            p=0.0
            for j in range(ite,nc):
                for i in rangel:
                    z=a[i,j]
                    if np.abs(z) > np.abs(p):
                        p=z
                        il=i
                        ic=j
            
            #case of singular matrix (redundant constraints)
            if np.abs(p) < epsilon:
                self.row_perm=self.vec_perm(range(0,nl),ipl,self.irk)
                self.col_perm=self.vec_perm(range(0,nc),ipc,self.irk)
                return
            
            # update the rank
            self.irk=ite+1 # on met le -1 car ca correspond au nombre de 'nombres' interessants trouve
#            ipl[ite]=il
            ipl.append(il) 
            ipc.append(ic)
#            ipc[ite]=ic
            
            #permutation of rows i and il
            
            if il>ite:
                for j in range(0,nc):
                    z=a[ite,j]
                    a[ite,j]=a[il,j]
                    a[il,j]=z
            
            #permutation of columns i and ic
            
            if ic>ite:
                for i in range(0,nl):
                    z=a[i,ite]
                    a[i,ite]=a[i,ic]
                    a[i,ic]=z
            
            #non pathological case:
            if ite == nl-1 or ite == nc-1:
                self.row_perm=self.vec_perm(range(0,nl),ipl,self.irk)
                self.col_perm=self.vec_perm(range(0,nc),ipc,self.irk)
                return
                #attention, on fait bien commencer les indices a zero car ce sont des indices de vecteur d'indices...

            #transformation of the sub-matrix
            it1=ite+1
            for i in range(it1,nl):
                z=a[i,ite]/p
                a[i,ite]=z
                for j in range(it1,nc):
                    a[i,j]=a[i,j]-z*a[ite,j]
                
                
        

    def vec_perm(self,b,ipm,irk):
        """ Make the permutation of the 'irk' first elements of the b vector with the 'ipm'
        permutation vector
        
        :param b: the vector to permute
        :param ipm: the permutation vector
        :param irk: the number of elements to permute in the b vector
        
        """
        a=b
        
        for i in range(0,irk):
            #Begin to 0 because this are indices vectors. They don't have a 0 at the first element
            i1=ipm[i]
            if i1 != i:
                z=a[i]
                a[i]=a[i1]
                a[i1]=z
        

        return a
        
             
